@@ -1,6 +1,6 @@
 # -*- coding: UTF-8 -*-
 
-# (c) Jérôme Laheurte 2015
+# (c) Jérôme Laheurte 2015-2018
 # See LICENSE.txt
 
 import six
@@ -13,7 +13,7 @@ from ptk.lexer import ProgressiveLexer, EOF, token
 from ptk.grammar import Grammar, Production, GrammarError
 # production is only imported so that client code doesn't have to import it from grammar
 from ptk.grammar import production # pylint: disable=W0611
-from ptk.utils import Singleton, callbackByName
+from ptk.utils import Singleton, callbackByName, memoize
 
 
 class ParseError(Exception):
@@ -24,7 +24,7 @@ class ParseError(Exception):
     """
     def __init__(self, grammar, tok, state):
         self.token = tok
-        super(ParseError, self).__init__(six.u('Unexpected token "%s" in state "%s"') % (tok.value, sorted(state)))
+        super(ParseError, self).__init__(six.u('Unexpected token "%s" (%s) in state "%s"') % (tok.value, tok.type, sorted(state)))
 
         self._expecting = set()
         for terminal in grammar.tokenTypes():
@@ -95,16 +95,15 @@ class _ResolveError(Exception):
 
 @functools.total_ordering
 class _Item(object):
+    __slots__ = ('production', 'dot', 'terminal', 'index', 'shouldReduce', 'expecting')
+
     def __init__(self, prod, dot, terminal):
         self.production = prod
         self.dot = dot
         self.terminal = terminal
-
-    def shouldReduce(self):
-        """
-        Returns True if the dot is in last position
-        """
-        return self.dot == len(self.production.right)
+        self.index = None
+        self.shouldReduce = self.dot == len(self.production.right)
+        self.expecting = None if self.shouldReduce else self.production.right[self.dot]
 
     def next(self):
         """
@@ -184,6 +183,9 @@ class LRParser(Grammar):
         super(LRParser, self).__init__()
         self._restartParser()
 
+    def rstack(self):
+        return reversed(self.__stack)
+
     def newToken(self, tok):
         try:
             for action, stack in self._processToken(tok):
@@ -191,7 +193,11 @@ class LRParser(Grammar):
                     break
         except _Accept as exc:
             self._restartParser()
-            self.newSentence(exc.result)
+            return self.newSentence(exc.result)
+
+    def currentLRState(self):
+        for item in self.__stack[-1].state:
+            return item.index
 
     def _processToken(self, tok):
         while True:
@@ -209,8 +215,8 @@ class LRParser(Grammar):
         raise NotImplementedError
 
     @classmethod
-    def _createProductionParser(cls, name, priority):
-        return ProductionParser(callbackByName(name), priority, cls)
+    def _createProductionParser(cls, name, priority, attrs):
+        return ProductionParser(callbackByName(name), priority, cls, attrs)
 
     @classmethod
     def _createReduceAction(cls, item):
@@ -264,12 +270,14 @@ class LRParser(Grammar):
             logger.debug('State %d', index)
             for item in sorted(state):
                 logger.debug('    %s', item)
+                item.index = index
+            cls.__lrstates__.append(sorted(state))
         logger.info('%d states.', len(states))
 
     @classmethod
     def __computeStates(cls, start):
-        allSyms = cls.tokenTypes() | cls.nonterminals()
-        goto = dict()
+        allSyms = list(cls.tokenTypes() | cls.nonterminals())
+        goto = list()
         cls._startState = frozenset([_Item(start, 0, EOF)])
         states = set([cls._startState])
         stack = [cls._startState]
@@ -278,31 +286,28 @@ class LRParser(Grammar):
             stateClosure = cls.__itemSetClosure(state)
             for symbol in allSyms:
                 # Compute goto(symbol, state)
-                nextState = set()
-                for item in stateClosure:
-                    if not item.shouldReduce() and item.production.right[item.dot] == symbol:
-                        nextState.add(item.next())
+                nextState = frozenset([item.next() for item in stateClosure if item.expecting == symbol])
                 if nextState:
-                    nextState = frozenset(nextState)
-                    goto[(state, symbol)] = nextState
+                    goto.append(((state, symbol), nextState))
                     if nextState not in states:
                         states.add(nextState)
                         stack.append(nextState)
-        return states, goto
+        return states, dict(goto)
 
     @classmethod
     def __computeActions(cls, states, goto):
         cls.__actions__ = dict()
         reachable = set()
+        tokenTypes = cls.tokenTypes()
         for state in states:
             for item in cls.__itemSetClosure(state):
-                if item.shouldReduce():
+                if item.shouldReduce:
                     action = cls._createReduceAction(item)
                     reachable.add(item.production.name)
                     cls.__addReduceAction(state, item.terminal, action)
                 else:
                     symbol = item.production.right[item.dot]
-                    if symbol in cls.tokenTypes():
+                    if symbol in tokenTypes:
                         cls.__addShiftAction(state, symbol, cls._createShiftAction(goto[(state, symbol)]))
         return reachable
 
@@ -396,11 +401,12 @@ class LRParser(Grammar):
         self.__stack = [_StackItem(self._startState, None)]
 
     @classmethod
+    @memoize
     def __itemSetClosure(cls, items):
         result = set(items)
         while True:
             prev = set(result)
-            for item in [item for item in result if not item.shouldReduce()]:
+            for item in [item for item in result if not item.shouldReduce]:
                 symbol = item.production.right[item.dot]
                 if symbol not in cls.tokenTypes():
                     terminals = cls.first(*tuple(item.production.right[item.dot + 1:] + [item.terminal]))
@@ -414,10 +420,11 @@ class LRParser(Grammar):
 
 class ProductionParser(LRParser, ProgressiveLexer): # pylint: disable=R0904
     # pylint: disable=C0111,C0103,R0201
-    def __init__(self, callback, priority, grammarClass): # pylint: disable=R0915
+    def __init__(self, callback, priority, grammarClass, attributes): # pylint: disable=R0915
         self.callback = callback
         self.priority = priority
         self.grammarClass = grammarClass
+        self.attributes = attributes
 
         super(ProductionParser, self).__init__()
 
@@ -425,6 +432,11 @@ class ProductionParser(LRParser, ProgressiveLexer): # pylint: disable=R0904
     def prepare(cls, **kwargs): # pylint: disable=R0915
         # Obviously cannot use @production here
 
+        # When mixing async and sync parsers in the same program this may be called twice,
+        # because AsyncProductionParser inherits from ProductionParser
+        if cls.productions():
+            return
+
         # DECL -> identifier "->" PRODS
         prod = Production('DECL', cls.DECL)
         prod.addSymbol('identifier', 'name')
@@ -669,7 +681,7 @@ class ProductionParser(LRParser, ProgressiveLexer): # pylint: disable=R0904
 
     def P2(self):
         # 'name' is replaced in newSentence()
-        return [Production(None, self.callback, priority=self.priority)]
+        return [Production(None, self.callback, priority=self.priority, attributes=self.attributes)]
 
     def SYMNAME1(self, identifier):
         return identifier
