@@ -1,6 +1,6 @@
 # -*- coding: UTF-8 -*-
 
-# (c) Jérôme Laheurte 2015
+# (c) Jérôme Laheurte 2015-2018
 # See LICENSE.txt
 
 """
@@ -11,6 +11,13 @@ import six
 import re
 import collections
 
+
+def intValue(char):
+    if six.PY3 and isinstance(char, int): # Byte
+        return int(bytes([char]))
+    return int(char)
+
+
 #===============================================================================
 # Regex objects
 
@@ -183,10 +190,6 @@ class RegularExpression(object):
         """
         Match a whole string
         """
-
-        if six.PY3 and isinstance(string, bytes):
-            string = [chr(c).encode('ascii') for c in string]
-
         self.start()
         try:
             for char in string:
@@ -229,7 +232,7 @@ class AnyCharacterClass(CharacterClass):
     """The ."""
 
     def __contains__(self, char):
-        return char not in [six.u('\n'), six.b('\n')]
+        return char not in [six.u('\n'), six.b('\n')[0]]
 
     def __eq__(self, other):
         return isinstance(other, AnyCharacterClass)
@@ -257,7 +260,7 @@ class RegexCharacterClass(CharacterClass): # pylint: disable=R0903
         return self is other # Because of cache
 
     def __contains__(self, char):
-        return self._rx.match(char) is not None
+        return self._rx.match(bytes([char]) if six.PY3 and isinstance(char, int) else char) is not None
 
 
 class LitteralCharacterClass(CharacterClass): # pylint: disable=R0903
@@ -297,11 +300,7 @@ class RegexTokenizer(object): # pylint: disable=R0903
     Token = collections.namedtuple('Token', ['type', 'value'])
 
     def __init__(self, regex):
-        if six.PY3 and isinstance(regex, bytes):
-            self._stack = [chr(c).encode('ascii') for c in reversed(regex)]
-        else:
-            self._stack = list(reversed(regex))
-
+        self._stack = list(reversed(regex))
         self._state = 0
         self._currentClass = None
         self._exponentValue = 0
@@ -313,7 +312,7 @@ class RegexTokenizer(object): # pylint: disable=R0903
             self.bbackslash = '\\'
         else:
             self.ubackslash = '\\'
-            self.bbackslash = '\\'.encode('ascii')
+            self.bbackslash = '\\'.encode('ascii')[0]
 
     def tokens(self):
         """
@@ -342,30 +341,30 @@ class RegexTokenizer(object): # pylint: disable=R0903
 
     def _state0(self, char, tokenList):
         # Normal state
-        if char in [six.u('*'), six.b('*')]:
+        if char in [six.u('*'), six.b('*')[0]]:
             tokenList.append(self.Token(self.TOK_EXPONENT, ExponentToken(0, None)))
-        elif char in [six.u('+'), six.b('+')]:
+        elif char in [six.u('+'), six.b('+')[0]]:
             tokenList.append(self.Token(self.TOK_EXPONENT, ExponentToken(1, None)))
-        elif char in [six.u('.'), six.b('.')]:
+        elif char in [six.u('.'), six.b('.')[0]]:
             tokenList.append(self.Token(self.TOK_CLASS, AnyCharacterClass()))
-        elif char in [six.u('('), six.b('(')]:
+        elif char in [six.u('('), six.b('(')[0]]:
             tokenList.append(self.Token(self.TOK_LPAREN, char))
-        elif char in [six.u(')'), six.b(')')]:
+        elif char in [six.u(')'), six.b(')')[0]]:
             tokenList.append(self.Token(self.TOK_RPAREN, char))
-        elif char in [six.u('|'), six.b('|')]:
+        elif char in [six.u('|'), six.b('|')[0]]:
             tokenList.append(self.Token(self.TOK_UNION, char))
         elif char == six.u('['):
             self._currentClass = six.StringIO()
             self._currentClass.write(char)
             return 2
-        elif char == six.b('['):
+        elif char == six.b('[')[0]:
             self._currentClass = six.BytesIO()
-            self._currentClass.write(char)
+            self._currentClass.write(bytes([char]) if six.PY3 else char)
             return 2
-        elif char in [six.u('{'), six.b('{')]:
+        elif char in [six.u('{'), six.b('{')[0]]:
             return 9
-        elif char in [six.u(']'), six.b(']'), six.u('}'), six.b('}')]:
-            raise TokenizeError('Unexpected token "%s"' % char)
+        elif char in [six.u(']'), six.b(']')[0], six.u('}'), six.b('}')[0]]:
+            raise TokenizeError('Unexpected token "%s"' % str(char))
         elif char in [self.ubackslash, self.bbackslash]:
             return 1
         else:
@@ -375,16 +374,20 @@ class RegexTokenizer(object): # pylint: disable=R0903
         # After a "\" in normal state
         if char in [six.u('d'), six.u('s'), six.u('w'), six.u('D'), six.u('S'), six.u('W')]:
             tokenList.append(self.Token(self.TOK_CLASS, RegexCharacterClass(self.ubackslash + char)))
-        elif char in [six.b('d'), six.b('s'), six.b('w'), six.b('D'), six.b('S'), six.b('W')]:
+        elif char in [six.b('d')[0], six.b('s')[0], six.b('w')[0], six.b('D')[0], six.b('S')[0], six.b('W')[0]]:
             tokenList.append(self.Token(self.TOK_CLASS, RegexCharacterClass(self.bbackslash + char)))
+        elif char == six.u('r'):
+            tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.u('\r'))))
         elif char == six.u('n'):
             tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.u('\n'))))
         elif char == six.u('t'):
             tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.u('\t'))))
-        elif char == six.b('n'):
-            tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.b('\n'))))
-        elif char == six.b('t'):
-            tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.b('\t'))))
+        elif char == six.b('r')[0]:
+            tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.b('\r')[0])))
+        elif char == six.b('n')[0]:
+            tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.b('\n')[0])))
+        elif char == six.b('t')[0]:
+            tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(six.b('\t')[0])))
         else:
             tokenList.append(self.Token(self.TOK_CLASS, LitteralCharacterClass(char)))
         return 0
@@ -395,64 +398,72 @@ class RegexTokenizer(object): # pylint: disable=R0903
         # In character class
         if char in [self.ubackslash, self.bbackslash]:
             return 3
-        if char in [six.u(']'), six.b(']')]:
-            self._currentClass.write(char)
+        if char in [six.u(']'), six.b(']')[0]]:
+            self._currentClass.write(bytes([char]) if six.PY3 and isinstance(char, int) else char)
             tokenList.append(self.Token(self.TOK_CLASS, RegexCharacterClass(self._currentClass.getvalue())))
             self._currentClass = None
             return 0
-        self._currentClass.write(char)
+        self._currentClass.write(bytes([char]) if six.PY3 and isinstance(char, int) else char)
 
     def _state3(self, char, tokenList): # pylint: disable=W0613
         # After "\" in character class
-        if six.PY2 and isinstance(char, str):
-            self._currentClass.write(self.bbackslash + char)
+        if (six.PY2 and isinstance(char, str)) or (six.PY3 and isinstance(char, int)):
+            self._currentClass.write(bytes([self.bbackslash]))
         else:
-            self._currentClass.write(self.ubackslash + char)
+            self._currentClass.write(self.ubackslash)
+        self._currentClass.write(bytes([char]) if six.PY3 and isinstance(char, int) else char)
         return 2
 
     # Exponent
 
     def _state9(self, char, tokenList): # pylint: disable=W0613
         # Start of exponent
-        if not char.isdigit():
+        try:
+            self._exponentValue = intValue(char)
+        except ValueError:
             raise InvalidExponentError('Exponent not starting with a number')
-        self._exponentValue = int(char)
         return 10
 
     def _state10(self, char, tokenList):
         # In exponent, computing start value
-        if char in [six.u('-'), six.b('-')]:
+        if char in [six.u('-'), six.b('-')[0]]:
             self._startExponent = self._exponentValue
             return 11
-        elif char in [six.u('}'), six.b('}')]:
+        elif char in [six.u('}'), six.b('}')[0]]:
             tokenList.append(self.Token(self.TOK_EXPONENT, ExponentToken(self._exponentValue, self._exponentValue)))
             return 0
-        elif char.isdigit():
-            self._exponentValue *= 10
-            self._exponentValue += int(char)
         else:
-            raise InvalidExponentError('Invalid character "%s"' % char)
+            try:
+                v = intValue(char)
+            except ValueError:
+                raise InvalidExponentError('Invalid character "%s"' % char)
+            self._exponentValue *= 10
+            self._exponentValue += v
 
     def _state11(self, char, tokenList): # pylint: disable=W0613
         # In exponent, expecting second term of interval
-        if char in [six.u('}'), six.b('}')]:
+        if char in [six.u('}'), six.b('}')[0]]:
             raise InvalidExponentError('Missing range end')
-        if not char.isdigit():
+        try:
+            v = intValue(char)
+        except ValueError:
             raise InvalidExponentError('Invalid character "%s"' % char)
-        self._exponentValue = int(char)
+        self._exponentValue = v
         return 12
 
     def _state12(self, char, tokenList):
         # In exponent, computing end value
-        if char in [six.u('}'), six.b('}')]:
+        if char in [six.u('}'), six.b('}')[0]]:
             if self._startExponent > self._exponentValue:
                 raise InvalidExponentError('Invalid exponent range %d-%d' % (self._startExponent, self._exponentValue))
             tokenList.append(self.Token(self.TOK_EXPONENT, ExponentToken(self._startExponent, self._exponentValue)))
             return 0
-        if not char.isdigit():
+        try:
+            v = intValue(char)
+        except ValueError:
             raise InvalidExponentError('Invalid character "%s"' % char)
         self._exponentValue *= 10
-        self._exponentValue += int(char)
+        self._exponentValue += v
 
 #===============================================================================
 # Parsing
@@ -475,7 +486,7 @@ class RegexParser(object):
         tokens = list(tokens)
         expr, pos = self._parse_E1(tokens, 0)
         if len(tokens) != pos:
-            raise RegexParseError('Unexpected token "%s"' % tokens[pos].value)
+            raise RegexParseError('Unexpected token "%s"' % str(tokens[pos].value))
         return expr
 
     def _parse_E1(self, tokens, pos):
@@ -523,7 +534,7 @@ class RegexParser(object):
             return expr, pos + 1
         elif tokens[pos].type == RegexTokenizer.TOK_CLASS:
             return self.klass(tokens[pos].value), pos + 1
-        raise RegexParseError('Unexpected token "%s"' % tokens[pos].value)
+        raise RegexParseError('Unexpected token "%s"' % str(tokens[pos].value))
 
     # Delegate
 
